@@ -2,7 +2,7 @@ use warnings;
 use strict;
 
 package Mail::IMAPClient;
-our $VERSION = '3.04';
+our $VERSION = '3.08';
 
 use Mail::IMAPClient::MessageSet;
 
@@ -346,10 +346,16 @@ sub separator
     return $self->{separators}{$target}
         if $self->{separators}{$target};
 
-    my $list = $self->list(undef, $target) || 'NO';
-    my $s    = $list =~ /^\*\s+LIST\s+\([^)]*\)\s+(\S+)/ ? $1 : '/';
-    $s       =~ s/^\"(.*?)\"$/$1/;
-    $self->{separators}{$target} = $s;
+    my $list = $self->list(undef, $target);
+    foreach my $line (@$list)
+    {   if($line =~ /^\*\s+LIST\s+\([^)]*\)\s+(\S+)/)
+        {   my $s = $1;
+            $s =~ s/^\"(.*?)\"$/$1/;
+            return $self->{separators}{$target} = $s;
+        }
+    }
+    $self->{separators}{$target} = '/';
+
 }
 
 sub sort
@@ -393,7 +399,6 @@ sub lsub
     defined $target    or $target = '*';
     $target = $self->Massage($target);
 
-    my $string      =
     $self->_imap_command( qq[LSUB "$reference" $target] )
          or return undef;
 
@@ -595,13 +600,12 @@ sub message_to_file
         binmode $handle; # For those of you who need something like this...
     }
 
-
     my $clear = $self->Clear;
     $self->Clear($clear)
         if $self->Count >= $clear && $clear > 0;
 
     my $peek   = $self->Peek ? '.PEEK' : '';
-    my $cmd    = $self->imap4rev1 ? "RFC822$peek" : "BODY$peek\[]";
+    my $cmd    = $self->imap4rev1 ? "BODY$peek\[]" : "RFC822$peek";
     my $uid    = $self->Uid ? "UID " : "";
     my $trans  = $self->Count($self->Count+1);
     my $string = "$trans ${uid}FETCH $msgs $cmd";
@@ -1300,8 +1304,9 @@ sub _read_line
     {   my $transno = $self->Transaction;
 
         if($timeout)
-        {   my @ready = $self->{_select}->can_read($timeout);
-            unless(@ready)
+        {   my $rvec = 0;
+            vec($rvec, fileno($self->Socket), 1) = 1;
+            unless(CORE::select($rvec, undef, $rvec, $timeout))
             {   $self->LastError("Tag $transno: Timeout after $timeout seconds"
                     . " waiting for data from server");
 
@@ -1566,11 +1571,12 @@ sub get_bodystructure
     }
     else
     {   $self->_debug("get_bodystructure: reassembling original response");
-        my $start = 0;
-        foreach my $o ($self->Results)
+        my $started = 0;
+        my $output  = '';
+        foreach my $o ($self->_transaction)
         {   next unless $self->_is_output_or_literal($o);
-            next unless $start or
-                $o->[DATA] =~ /BODYSTRUCTURE \(/i and ++$start; # Hi, vi! ;-)
+            $started++ if $o->[DATA] =~ /BODYSTRUCTURE \(/i; ; # Hi, vi! ;-)
+            $started or next;
 
             if(length $output && $self->_is_literal($o) )
             {   my $data = $o->[DATA];
@@ -1598,24 +1604,30 @@ sub get_envelope
         return undef;
     }
 
-    my @out = $self->fetch($msg,"ENVELOPE");
+    my @out = $self->fetch($msg, 'ENVELOPE');
     my $bs = "";
-    my $output = first { /ENVELOPE \(/i } @out;    # Wee! ;-)
+    my $output = first { /ENVELOPE \(/i } @out;    # vi ;-)
+
+    unless($output)
+    {   $self->LastError("Unable to use get_envelope: $@");
+        return undef;
+    }
+
     if($output =~ /\r\n$/ )
     {   eval { $bs = Mail::IMAPClient::BodyStructure::Envelope->new($output) };
     }
     else
     {   $self->_debug("get_envelope: reassembling original response");
-        my $start = 0;
-        foreach my $o ($self->Results)
+        my $started = 0;
+        $output = '';
+        foreach my $o ($self->_transaction)
         {   next unless $self->_is_output_or_literal($o);
             $self->_debug("o->[DATA] is $o->[DATA]");
 
-            next unless $start or
-                $o->[DATA] =~ /ENVELOPE \(/i and ++$start;
-                # Hi, vi! ;-)
+            $started++ if $o->[DATA] =~ /ENVELOPE \(/i; # Hi, vi! ;-)
+            $started or next;
 
-            if ( length($output) and $self->_is_literal($o) ) {
+            if(length($output) && $self->_is_literal($o) ) {
                 my $data = $o->[DATA];
                 $data =~ s/"/\\"/g;
                 $data =~ s/\(/\\\(/g;
@@ -1721,12 +1733,12 @@ sub store
     wantarray ? $self->History : $self->Results;
 }
 
-sub _imap_folder_command($$)
+sub _imap_folder_command($$@)
 {   my ($self, $command) = (shift, shift);
     delete $self->{Folders};
     my $folder = $self->Massage(shift);
 
-    $self->_imap_command("$command $folder")
+    $self->_imap_command(join ' ', $command, $folder, @_)
         or return;
 
     wantarray ? $self->History : $self->Results;
@@ -1735,7 +1747,7 @@ sub _imap_folder_command($$)
 sub subscribe($)   { $_[0]->_imap_folder_command(SUBSCRIBE   => $_[1]) }
 sub unsubscribe($) { $_[0]->_imap_folder_command(UNSUBSCRIBE => $_[1]) }
 sub delete($)      { $_[0]->_imap_folder_command(DELETE      => $_[1]) }
-sub create($)      { $_[0]->_imap_folder_command(CREATE      => $_[1]) }
+sub create($)      { my $self=shift; $self->_imap_folder_command(CREATE => @_)}
 
 # rfc2086
 sub myrights($)    { $_[0]->_imap_folder_command(MYRIGHTS    => $_[1]) }
@@ -1750,20 +1762,18 @@ sub close
 
 sub expunge
 {   my ($self, $folder) = @_;
-    defined $folder
-        or return;
 
-    my $old = $self->Folder;
-    if(defined $old && $folder eq $old)
+    my $old = $self->Folder || '';
+    if(defined $folder && $folder eq $old)
+    {   $self->_imap_command('EXPUNGE')
+            or return undef;
+    }
+    else
     {   $self->select($folder);
         my $succ = $self->_imap_command('EXPUNGE');
         $self->select($old);
         $succ or return undef;
     }
-    else
-    {   $self->_imap_command('EXPUNGE')
-            or return undef;
-    }
 
     wantarray ? $self->History : $self->Results;
 }
@@ -2298,6 +2308,7 @@ sub append_string($$$;$$)
         if $self->Count >= $clear and $clear > 0;
 
     my $count  = $self->Count($self->Count+1);
+    $text =~ s/\r?\n/\r\n/g;
 
     my $command = "$count APPEND $folder " . ($flags ? "$flags " : "") .
         ($date ? "$date " : "") .  "{" . length($text) . "}\r\n";
@@ -2494,10 +2505,12 @@ sub authenticate
     until($code)
     {   my $output = $self->_read_line or return undef;
         foreach my $o (@$output)
-        {   $self->_record($count,$o);
-            $code = $o->[DATA] =~ /^\+\s+(.*)$/ ? $1 : undef;
+        {   $self->_record($count, $o);
+            $code = $o->[DATA] =~ /^\+\s+(\S+)\s*$/ ? $1
+                  : $o->[DATA] =~ /^\+\s*$/        ? 'OK'
+                  : undef;
 
-            if ($o->[DATA] =~ /^\*\s+BYE/)
+            if($o->[DATA] =~ /^\*\s+BYE/)
             {   $self->State(Unconnected);
                 return undef;
             }
@@ -2563,12 +2576,12 @@ sub authenticate
         return undef;
     }
 
-    undef $code = $scheme eq 'PLAIN' ? 'OK' : undef;
+    undef $code;
     until($code)
     {   my $output = $self->_read_line or return undef;
         foreach my $o (@$output)
         {   $self->_record($count, $o);
-            $code = $o->[DATA] =~ /^\+\s+(.*)$/ ? $1 : undef;
+            $code = $o->[DATA] =~ /^\+\s+(.*?)\s*$/ ? $1 : undef;
 
             if($code)
             {   unless($self->_send_line($response->($code, $self)))
@@ -2706,10 +2719,23 @@ sub getquota
     $self->_imap_command("GETQUOTA $who") ? $self->Results : undef;
 }
 
+# usage: $imap->setquota($folder, storage => 512)
+sub setquota(@)
+{   my ($self, $what) = (shift, shift);
+    my $who = $what ? $self->Massage($what) : "user/$self->{User}";
+    my @limits;
+    while(@_)
+    {   my $key = uc shift @_;
+        push @limits, $key => shift @_;
+    }
+    local $" = ' ';
+    $self->_imap_command("SETQUOTA $who (@limits)") ? $self->Results : undef;
+}
+
 sub quota
 {   my $self = shift;
     my $what = shift || "INBOX";
-    $self->_imap_command("GETQUOTA $what") || $self->getquotaroot($what);
+    $self->_imap_command("GETQUOTA $what") or $self->getquotaroot($what);
     (map { /.*STORAGE\s+\d+\s+(\d+).*\n$/ ? $1 : () } $self->Results)[0];
 }
 
